import java.lang.reflect.*;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * Aspect has a name and a table for pointcuts associated with it.
 *
 */
public class Aspect {
  private String _aspect_name;
  public Hashtable _pointcuts = new Hashtable();

  /**
   * New Aspect object must always be created with a name provided.
   */
  public Aspect(String name) {
    _aspect_name = name;
  }

  /**
   * getName() returns the name of the aspect.
   */
  public String getName() {
    return _aspect_name;
  }

  /**
   * Returns the pointcut with the pc_name
   * from the pointcuts table.
   * Returns null if the pointcut does not exist.
   */
  public Pointcut getPointcut(String pc_name) {
     return (Pointcut) _pointcuts.get(pc_name);
  }

  /**
   * Adds a pointcut with the given pointcut_name to the pointcut table.
   * If the pointcut is a new one, it is added to the pointcut table.
   *
   * Returns the pointcut with the given pointcut_name.
   */
  public Pointcut addPointcut(String pointcut_name) {
    if (!_pointcuts.containsKey(pointcut_name)) {
       _pointcuts.put(pointcut_name, new Pointcut(pointcut_name));
       System.out.println("Pointcut added: " + pointcut_name);
    }
    return (Pointcut) _pointcuts.get(pointcut_name);
  }

  /**
   * An advice for the given pointcut and method is
   * added to the current aspect object.
   *
   * If the pointcut with the given pointcut_name is not registered,
   * a new one will be created and added to the pointcut table inside
   * the aspect.
   *
   * If the reference to an advice for the given method_name
   * inside the given pointcut of the given aspect is already there,
   * the reference will be overwritten with the new advice reference.
   *
   */
  public void addPointcut(String pointcut_name, String method_name,
                          AdviceIF advice){
       Pointcut p = this.addPointcut(pointcut_name);
       p.addAdvice(method_name, advice);
  }

  /**
   * Removes a pointcut of the given name
   * from the pointcut table of this Aspect.
   */
  public void removePointcut(String pointcut_name){
    if (_pointcuts.containsKey(pointcut_name)) {
        _pointcuts.remove(pointcut_name);
    }
  }

  /**
   * beforeAdvice() is used by AspectRepository's beforeAdvice() method.
   * obj is a reference to the shared object, for example, bounded buffer.
   * m is the shared object's method to be invoked if all the beforeAdvice() is successful.
   * args is the array that contains all the arguments passed to the method m.
   *
   */
  public int beforeAdvice(Object obj, Method m, Object[] args){
    //System.out.println("Aspect.beforeAdvice()");
    if (_pointcuts.containsKey("DEFAULT")) {
      Pointcut p = (Pointcut) _pointcuts.get("DEFAULT");
      return p.beforeAdvice(obj, m, args);
    }
    return AspectRepository.RESUME;
  }

  /**
   * afterAdvice() is used by AspectRepository's afterAdvice() method.
   * obj is a reference to the shared object, for example, bounded buffer.
   * m is the shared object's method to be invoked if all the beforeAdvice() is successful.
   * args is the array that contains all the arguments passed to the method m.
   * result is the object returned by the method m of the shared object obj.
   */
  public Object afterAdvice(Object obj, Method m, Object[] args, Object result){
    if (_pointcuts.containsKey("DEFAULT")) {
      Pointcut p = (Pointcut) _pointcuts.get("DEFAULT");
      return p.afterAdvice(obj, m, args, result);
    }
    return null;
  }
}
