import java.util.Vector;
import java.lang.reflect.*;
import java.util.Observable;
import java.util.Observer;
import java.util.Hashtable;
import java.util.Enumeration;
import javax.swing.*;

public class AspectRepository
  extends Observable {

  /* Constants for the return value of AdviceIF.beforeAdvice().
   * ABORT : current thread must be aborted.
   * BLOCK : current thread must be blocked.
   * RESUME: current thread can go on to the next operation.
   */
  public static final int ABORT  = 1, BLOCK  = 2, RESUME = 3;

  /*
   * _history is a queue of buffer access history.
   * Each queue element is a HistoryElem object.
   */
  private HistoryQueue _history = new HistoryQueue();

  /*
   * Class name of the current scheduling policy
   */
  private String currPolicyName = "Adv_PreferFIFO";

  /*
   * AspectTable contains all the aspects
   * to be used in this AspectRepository.
   */
  private AspectTable aspectTbl = new AspectTable();



  public AspectRepository() {
  }

  public String getCurrPolicyName() { return currPolicyName; }

  public synchronized void changePolicy(String className) {
    if (className.equals(currPolicyName)) return;

    // load the new policy
    AdviceIF newPolicyAdvice = loadAdvice(className);
    if (newPolicyAdvice == null) return;
    this.addAspect("SCHED", "DEFAULT", "DEFAULT", newPolicyAdvice);
    currPolicyName = className;
  }

  public void addAspect(String aspect_name, String pointcut_name,
                        String method_name, AdviceIF advice){
    aspectTbl.addAspect(aspect_name, pointcut_name, method_name, advice);
  }

  public void removeAspect(String aspect_name, String pointcut_name,
                           String method_name){
    aspectTbl.removeAspect(aspect_name, pointcut_name, method_name);
  }

  /*----------------------------------------------------------------
   * When a task requests entry to the bounded buffer, it must wait
   * outside it until it is admitted by the AspectRepository.
   * It can be accomplished by defining the beforeAdvice() method
   * as synchronized method.
   * Once a task is inside the beforeAdvice(), it is queued and
   * awaits entry. To guarantee the mutual exclusion, only one
   * task is allowed entrance at a time.
   * Once a task has completed its operation on the bounded buffer,
   * it must reevaluate all the waiting at their respective queues.
   * If there are multiple entries within AspectRepository, we have
   * three options to choose from:
   *    Preference to Put
   *    Preference to Get
   *    Preference to None (FIFO)
   *----------------------------------------------------------------
   */
  private boolean occupied = false;  // A task is accessing bounded buffer.
  //private Task activeTask = null;    // The current task that is accessing the buffer
  private Task taskToBeAwakened = null;

  //public Task getActiveTask() { return activeTask; }

  public synchronized void beforeAdvice(Object obj, Method m, Object[] args) {
    Task task = new Task(m.getName()); // init task info
    int i, rc;
    Aspect a;
    Preference policyAdvice;
    boolean mustWait = false;

    /**
     * Invoke every beforeAdvice() method of each aspect.
     */
    for(i = 0; i < aspectTbl.size(); i++) {
      a = aspectTbl.getAspect(i);
      //System.out.println(Thread.currentThread().getName() + " " + m.getName() +
      //                   " AR.beforeAdvice: " + a.getName());
      mustWait = false;
      if (occupied)
        mustWait = true;
      else {
        rc = a.beforeAdvice(obj, m, args);
        if (BLOCK == rc)
          mustWait = true;
      }

      while (mustWait) {
        policyAdvice = (Preference)
          aspectTbl.getAdvice("SCHED", "DEFAULT", "DEFAULT");

        //System.out.println(Thread.currentThread().getName() + " " + m.getName() +
        //                 " Waits");
        policyAdvice.enqueue(task);
        try { this.wait(); }        //------------- WAIT
        catch(InterruptedException iex) { /* Ignore iex */ }
        policyAdvice.dequeue(task);
        //System.out.println(Thread.currentThread().getName() + " " + m.getName() +
        //                 " Notified");
        if (task.equals(taskToBeAwakened)) {

          mustWait = false; // Exit this while loop and
          i = -1; // Force all the aspects to be checked again.
          //System.out.println("AR.beforeAdvice: task awakened : " +
          //                   task.getName());
        }
      }
    }
    occupied = true;  // Only one task at a time is allowed
                      // entrance to the bounded buffer.

    //System.out.println(Thread.currentThread().getName() + " " + m.getName() +
    //                   " Enters buffer");
  }

  /*--------------------------------------------------
   *--------------------------------------------------
   */
  public synchronized void afterAdvice(Object obj, Method m, Object[] args, Object result) {
    int i, rc;
    Aspect a;

    /**
     * Invoke every afterAdvice() method of each aspect in reverse order.
     */
    for(i = aspectTbl.size() - 1; i >= 0 ; i--) {
      a = aspectTbl.getAspect(i);
      //System.out.println(Thread.currentThread().getName() + " " + m.getName() +
      //                   " AR.afterAdvice: " + a.getName());

      if (a.getName().equals("SCHED")) {
        // Pick a task to be unblocked.
        taskToBeAwakened = (Task) a.afterAdvice(obj, m, args, result); // obj is bounded buffer
        if (taskToBeAwakened != null) {
          //System.out.println("AR.afterAdvice: task to be awakened : " +
          //                   taskToBeAwakened.getName());
          this.notifyAll();                 //  <--------------- Wake up.
        } else {
            //System.out.println("AR.afterAdvice: no notifications to waiting tasks");
        }

      } else {
        a.afterAdvice(obj, m, args, result);
      }
    }
    occupied = false;

    // Update the screen display by notifying Observers of the changes
    Preference policyAdvice =
      (Preference) aspectTbl.getAdvice("SCHED", "DEFAULT", "DEFAULT");

    _history.enqueue(policyAdvice.getQsize("put"),
                     policyAdvice.getQsize("get"),
                     ((BoundedBuffer) obj).getCnt(),
                     policyAdvice.getPolicyName());
    setChanged();
    notifyObservers(_history);
  }

  public void clearHistory() {
    _history.clear();
  }

  public Vector getAllPointCutInfos() { // returns a Vector of ...
    return aspectTbl.getAllPointCutInfos();
  }

  public AdviceIF loadAdvice(String className) {
    /**
     * Loads an advice from the specified file.
     * Success returns the advice loaded.
     * Error returns null.
     */
    AdviceIF advice = null;

    try {
      Class myClass = Class.forName(className);
      advice = (AdviceIF) myClass.newInstance();
      System.out.println("Loading Advice <" + className + "> in AspectRepository");
    } catch (Exception ex) {
      System.out.println("Error loading Advice <" + className + "> in AspectRepository");
    }
    return advice;
  }


  /*
   * Vector of class names of available scheduling policies
   */
  private Vector _policies = new Vector();

  public void addPolicy(String policyName) {
    /**
     * Adds policyName to the _policies Vector.
     */
    if (!_policies.contains(policyName))
      _policies.add(policyName);
  }

  /**
   * Removes policyName from the _policies Vector.
   */
  public void removePolicy(String policyName) {
    if (_policies.contains(policyName)) {
      int i = _policies.indexOf(policyName);
      _policies.remove(i);
      // JOptionPane.showMessageDialog(null, "removed " + policyName );
    }
  }

  /**
   *    returns a Vector of class names of available policies
   */
  public Vector getAllPolicyInfo() {
    return _policies;
  }
}
