import java.util.Vector;
import java.lang.reflect.*;
import java.util.Enumeration;

/**
 * AspectTable contains all the aspects to be used in this system.
 * AspectTable will only be contained in the AspectRepository.
 * AspectRepository delegates the resposibility of maintaining the
 * Aspects, Pointcuts, and JoinPoints to the AspectTable.
 */
public class AspectTable {
  private Vector _aspects = new Vector(); // contains instances of Aspect.

  /**
   * Returns the number of aspects stored in this table.
   */
  public int size() {

     if (_aspects == null) return 0;
     return _aspects.size();
  }

    /**
     * Checks if an aspect with the given name is already registered
     * in the aspect table.
     * Returns true if it is, otherwise false.
     */
  public boolean hasAspect(String aspect_name) {
     int i;

     if (_aspects == null) return false;
     if (_aspects.isEmpty()) return false;

     for(i = 0; i < _aspects.size(); i++) {
       AspectIF a = (AspectIF) _aspects.get(i);
       if (a.getName().equals(aspect_name))
         return true;
     }
     return false;
  }

  /**
   * Returns a reference to the aspect with the aspect_name
   * from the aspect table.
   * Returns null if the aspect does not exist.
   */
  public Aspect getAspect(String aspect_name) {
     int i;

     if (_aspects == null) return null;
     if (_aspects.isEmpty()) return null;

     for(i = 0; i < _aspects.size(); i++) {
       Aspect a = (Aspect) _aspects.get(i);
       if (a.getName().equals(aspect_name))
         return a;
     }
     return null;
  }

    /**
     * Returns a reference to the aspect in index position
     * from the aspect table.
     * Returns null if the index value is illegal.
     */
  public Aspect getAspect(int index) {
     int i;

     if (_aspects == null) return null;
     if (_aspects.isEmpty()) return null;
     if (_aspects.size() <= index || index < 0) return null;

     return (Aspect) _aspects.get(index);
  }

  /*
   * Adds an aspect with the given aspect_name to the aspect table.
   * If the aspect is a new one, an aspect is appended to the end
   * of the aspect table.
   *
   * Reference to the aspect is returned.
   */
  private Aspect addAspect(String aspect_name) {
     Aspect a = this.getAspect(aspect_name);
     if (a == null) {
       a = new Aspect(aspect_name);
       _aspects.add(a);
     }
     return a;
  }

  /**
   * An advice for the given aspect, pointcut, and method is
   * added to the system.
   *
   * If the aspect with the given aspect_name does not exist,
   * a new one will be created and added to the aspect table.
   *
   * If the pointcut with the given pointcut_name is not registered,
   * a new one will be created and added to the pointcut table inside
   * the given aspect.
   *
   * If the reference to an advice for the given method_name
   * inside the given pointcut of the given aspect is already there,
   * the reference will be overwritten with the new advice reference.
   *
   * Returns true.
   */
  public boolean addAspect(String aspect_name, String pointcut_name,
                       String method_name, AdviceIF advice){


     Aspect a = this.addAspect(aspect_name);
     Pointcut p = a.addPointcut(pointcut_name);
     p.addAdvice(method_name, advice);
     return true;
  }

  /*
   * Removes an aspect of the given name from the aspect table.
   */
  private void removeAspect(String aspect_name){
     Aspect a = this.getAspect(aspect_name);
     if (a != null)
       _aspects.remove(a);
  }

  /**
   * Removes an advice for the given method name of the given pointcut
   * from the given aspect.
   */
  public void removeAspect(String aspect_name, String pointcut_name,
                           String method_name){
     System.out.println("Trying to Remove the advice for " + aspect_name + ", "
                       + pointcut_name + ", " + method_name);

     Aspect a = this.getAspect(aspect_name);
     if (a == null) {
       System.out.println("No such aspect found : " + aspect_name);
       return;
    }

     Pointcut p = a.getPointcut(pointcut_name);
     if (p == null) {
       System.out.println("No such pointcut found : " + pointcut_name);
       return;
    }

     AdviceIF adv = p.getAdvice(method_name);
     if (adv == null) {
       System.out.println("No such method found : " + method_name);
       return;
     }
     p.removeAdvice(method_name);
     System.out.println("Removed the advice for " + aspect_name + ", "
                       + pointcut_name + ", " + method_name);
  }



  /**
   * Returns a reference to the advice at the given aspect, pointcut, and
   * the join point.
   * Returns null if such advice does not exist.
   */
  public AdviceIF getAdvice(String aspect_name, String pointcut_name, String method_name) {
     int i;

     if (_aspects == null) return null;
     if (_aspects.isEmpty()) return null;

     for(i = 0; i < _aspects.size(); i++) {
       Aspect a = (Aspect) _aspects.get(i);
       if (a.getName().equals(aspect_name)) {
         Pointcut p = (Pointcut) a.getPointcut(pointcut_name);
         if (p == null)
           return null;
         else
           return p.getAdvice(method_name);
       }
     }
     return null;
  }


  /**
   * Returns a Vector of registered advices.
   * Each element of the Vector is a String that contains the
   * aspect name, pointcut name, join points name, and advice class name.
   */
  public Vector getAllPointCutInfos() {
    Vector _all_infos = new Vector();
    int i;

    if (_aspects == null) return _all_infos;
    if (_aspects.isEmpty()) return _all_infos;

    for(i = 0; i < _aspects.size(); i++) {
      Aspect asp = (Aspect) _aspects.get(i);
      for (Enumeration eP = asp._pointcuts.elements() ; eP.hasMoreElements() ;) {
        Pointcut ptc = (Pointcut) eP.nextElement();
        for (Enumeration eK = ptc._advices.keys(); eK.hasMoreElements() ;) {
          String method_name = (String) eK.nextElement();

          if (method_name != null) {
            AdviceIF adv = (AdviceIF) ptc._advices.get(method_name);
            if (adv != null) {
              String ss = asp.getName() + ", " + ptc.getName() + ","
                        + method_name + ", " + adv.getClass().getName();
              _all_infos.add(ss);
            }
          }
        }
      }
    }
    return _all_infos;
  }
}
