import java.lang.reflect.*;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.io.*;

/**
 * This class can convert from collections of value objects to
 * Java objects. The converter uses a hashtable to prevent multiple
 * objects being created from multiple references to the one reference
 * or array value.
 */

class ValueConversionException extends Exception {

  public ValueConversionException () {
    super();
  }

  public ValueConversionException (String message) {
    super(message);
  }
}


/**
 * This interface defines character codes which uniquely identify
 * each type of value which the persistence mechanism can deal
 * with.
 */

interface ValueCodes {

  public static final char ARRAY = 'a';
  public static final char BOOLEAN = 'b';
  public static final char BYTE = 'y';
  public static final char CHAR = 'c';
  public static final char DOUBLE = 'd';
  public static final char FLOAT = 'f';
  public static final char INT = 'i';
  public static final char LONG = 'l';
  public static final char NULL = 'n';
  public static final char REFERENCE = 'r';
  public static final char SHORT = 's';
  public static final char STRING = 't';
}


/**
 * Represents a specific element in an array. Used to store an array
 * reference and an index in a hashtable.
 */

class ArrayElement {

  private Object array;
  private int index;

  public ArrayElement (Object array, int index) {

    this.array = array;
    this.index = index;
  }

  public void setElement (Object ref) throws Exception {
    Array.set(array, index, ref);
  }
}

class ArrayValue implements Value {

  private Vector elements;
  private String elementType;
  private int length;

  public ArrayValue () {

    elementType = null;
    length = 0;
    elements = new Vector();
  }

  public ArrayValue (String elementType, int length) {

    this.elementType = elementType;
    this.length = length;
    elements = new Vector();
  }

  public String getElementType () {
    return elementType;
  }

  public int getLength () {
    return length;
  }

  public void addElement (Value v) {

    if (elements != null)
      elements.addElement(v);
  }

  public int getSize () {

    if (elements == null)
      return 0;
    else
      return elements.size();
  }


  public Value getValueAt (int index) {

    if (elements == null)
      return null;

    if (index >= 0 && index < length)
      return (Value) elements.elementAt(index);
    else
      return null;
  }

  public void setValueAt (int index, Value newValue) {

    if (elements == null || index < 0 || index >= length)
      return;

    elements.setElementAt(newValue, index);
  }

  public Enumeration getElements () {

    if (elements == null)
      return null;
    else
      return elements.elements();
  }

  public String toString () {

    if (elements == null) {
      return elementType + "[]";
    }

    StringBuffer buffer = new StringBuffer();

    buffer.append(elementType);
    buffer.append('[');

    Enumeration e = elements.elements();

    if (e.hasMoreElements())
      buffer.append(e.nextElement());

    while (e.hasMoreElements()) {

      buffer.append(',');
      buffer.append(e.nextElement());
    }

    buffer.append(']');

    return buffer.toString();
  }
}

/**
 * This class represents a primitive boolean value.
 */

class BooleanValue implements Value {

  private boolean value;

  public BooleanValue () {
  }

  public BooleanValue (boolean value) {
    this.value = value;
  }

  public boolean booleanValue () {
    return value;
  }

  public String toString () {
    return String.valueOf(value);
  }
}

/**
 * This class represents a primitive byte value.
 */

class ByteValue implements Value {

  private byte value;

  public ByteValue () {
  }

  public ByteValue (byte value) {
    this.value = value;
  }

  public byte byteValue () {
    return value;
  }

  public String toString () {
    return String.valueOf(value);
  }
}

/**
 * This class represents a primitive character value.
 */

class CharValue implements Value {

  private char value;

  public CharValue () {
  }

  public CharValue (char value) {
    this.value = value;
  }

  public char charValue () {
    return value;
  }

  public String toString () {
    return  "'" + value + "'";
  }
}

/**
 * This class represents a primitive double value.
 */

class DoubleValue implements Value {

  private double value;

  public DoubleValue () {
  }

  public DoubleValue (double value) {
    this.value = value;
  }

  public double doubleValue () {
    return value;
  }
 
  public String toString () {
    return String.valueOf(value);
  } 
}

/**
 * This class represents a primitive float value.
 */

class FloatValue implements Value {

  private float value;

  public FloatValue () {
  }

  public FloatValue (float value) {
    this.value = value;
  }

  public float floatValue () {
    return value;
  }
  
  public String toString () {
    return String.valueOf(value);
  }
}

/**
 * This class represents a primitive integer value.
 */

class IntValue implements Value {

  private int value;

  public IntValue () {
  }

  public IntValue (int value) {
    this.value = value;
  }

  public int intValue () {
    return value;
  }
  
  public String toString () {
    return String.valueOf(value);
  }
}

/**
 * This class represents a primitive long value.
 */

class LongValue implements Value {

  private long value;

  public LongValue () {
  }

  public LongValue (long value) {
    this.value = value;
  }

  public long longValue () {
    return value;
  }

  public String toString () {
    return String.valueOf(value);
  }
}

/**
 * This class represents the null value. Null value objects are
 * placeholders which are used to indicate null references, instead
 * of an empty instance of another kind of value.
 */

class NullValue implements Value {

  public NullValue () {
  }

  public String toString () {
    return "null";
  }
}

/**
 * This class can convert from Java objects, arrays and primitive values
 * to collections of value objects.
 */

class ObjectConverter {



  private Hashtable refs;
  private Value value;
  private Object rootObject;
  private Value dummyRootValue;



  public ObjectConverter (Object object) throws ValueConversionException {

    refs = new Hashtable();
    value = null;

    dummyRootValue = new NullValue();

    try {

      if (object == null)
        value = new NullValue();
      else
        value = convertObjectToValue(object);

      convertDummyRootValue(value, value);

    } catch (Exception e) {

      e.printStackTrace();
      throw new ValueConversionException(e.getMessage());
    }
  }



  public ObjectConverter (boolean b) {
    value = new BooleanValue(b);
  }



  public ObjectConverter (byte b) {
    value = new ByteValue(b);
  }



  public ObjectConverter (char c) {
    value = new CharValue(c);
  }



  public ObjectConverter (double d) {
    value = new DoubleValue(d);
  }



  public ObjectConverter (float f) {
    value = new FloatValue(f);
  }



  public ObjectConverter (int i) {
    value = new IntValue(i);
  }



  public ObjectConverter (long l) {
    value = new LongValue(l);
  }



  public ObjectConverter (short s) {
    value = new ShortValue(s);
  }



  public Value getValue () throws ValueConversionException {

    if (value == null)
      throw new ValueConversionException("unsuccessful conversion");
    else 
      return value;
  }



  private Value convertObjectToValue (Object object) throws Exception {

    if (object == null)
      return new NullValue();

    if (object == rootObject)
      return dummyRootValue;

    // Set the root object the first time this method is called.

    if (rootObject == null)
       rootObject = object;

    if (refs.containsKey(object))
      return (Value) refs.get(object);

    Class c = object.getClass();
    Value v = null;

    if (c == java.lang.String.class)
      v = convertStringToValue(object);

    else if (c.isArray())
      v = convertArrayToValue(object);

    else
      v = convertReferenceToValue(object);

    if (v != null)
      refs.put(object, v);

    return v;
  }



  private Value convertStringToValue (Object object) throws Exception {
    return new StringValue((String) object);
  }



  private Value convertArrayToValue (Object array) throws Exception {

    Class type = array.getClass();
    Class elementType = type.getComponentType();
    String elementTypeName = elementType.getName();

    if (elementType == null)
      throw new ValueConversionException("Field is not an array: " + array);

    if (elementType == Boolean.TYPE)
      return convertBooleanArrayToValue(array, elementTypeName);

    if (elementType == Byte.TYPE)
      return convertByteArrayToValue(array, elementTypeName);

    if (elementType == Character.TYPE)
      return convertCharArrayToValue(array, elementTypeName);

    if (elementType == Double.TYPE)
      return convertDoubleArrayToValue(array, elementTypeName);

    if (elementType == Float.TYPE)
      return convertFloatArrayToValue(array, elementTypeName);

    if (elementType == Integer.TYPE)
      return convertIntArrayToValue(array, elementTypeName);

    if (elementType == Long.TYPE)
      return convertLongArrayToValue(array, elementTypeName);

    if (elementType == Short.TYPE)
      return convertShortArrayToValue(array, elementTypeName);

    if (elementType == java.lang.String.class)
      return convertStringArrayToValue(array, elementTypeName);

    if (elementType.isArray())
      return convertArrayArrayToValue(array, elementTypeName);

    return convertReferenceArrayToValue(array, elementTypeName);
  }



  private Value convertBooleanArrayToValue (Object array, String elementType)
   throws Exception {

    int length = Array.getLength(array);
    ArrayValue value = new ArrayValue(elementType, length);

    for (int i = 0; i < length; i++) {

      Value element = new BooleanValue(Array.getBoolean(array, i));
      value.addElement(element);
    }

    return value;
  }



  private Value convertByteArrayToValue (Object array, String elementType)
   throws Exception {

    int length = Array.getLength(array);
    ArrayValue value = new ArrayValue(elementType, length);

    for (int i = 0; i < length; i++) {

      Value element = new ByteValue(Array.getByte(array, i));
      value.addElement(element);
    }

    return value;
  }



  private Value convertCharArrayToValue (Object array, String elementType)
   throws Exception {

    int length = Array.getLength(array);
    ArrayValue value = new ArrayValue(elementType, length);

    for (int i = 0; i < length; i++) {

      Value element = new CharValue(Array.getChar(array, i));
      value.addElement(element);
    }

    return value;
  }



  private Value convertDoubleArrayToValue (Object array, String elementType)
   throws Exception {

    int length = Array.getLength(array);
    ArrayValue value = new ArrayValue(elementType, length);

    for (int i = 0; i < length; i++) {

      Value element = new DoubleValue(Array.getDouble(array, i));
      value.addElement(element);
    }

    return value;
  }



  private Value convertFloatArrayToValue (Object array, String elementType)
   throws Exception {

    int length = Array.getLength(array);
    ArrayValue value = new ArrayValue(elementType, length);

    for (int i = 0; i < length; i++) {

      Value element = new FloatValue(Array.getFloat(array, i));
      value.addElement(element);
    }

    return value;
  }



  private Value convertIntArrayToValue (Object array, String elementType)
   throws Exception {

    int length = Array.getLength(array);
    ArrayValue value = new ArrayValue(elementType, length);

    for (int i = 0; i < length; i++) {

      Value element = new IntValue(Array.getInt(array, i));
      value.addElement(element);
    }

    return value;
  }



  private Value convertLongArrayToValue (Object array, String elementType)
   throws Exception {

    int length = Array.getLength(array);
    ArrayValue value = new ArrayValue(elementType, length);

    for (int i = 0; i < length; i++) {

      Value element = new LongValue(Array.getLong(array, i));
      value.addElement(element);
    }

    return value;
  }



  private Value convertShortArrayToValue (Object array, String elementType)
   throws Exception {

    int length = Array.getLength(array);
    ArrayValue value = new ArrayValue(elementType, length);

    for (int i = 0; i < length; i++) {

      Value element = new ShortValue(Array.getShort(array, i));
      value.addElement(element);
    }

    return value;
  }



  private Value convertStringArrayToValue (Object array, String elementType)
   throws Exception {

    int length = Array.getLength(array);
    ArrayValue value = new ArrayValue(elementType, length);

    for (int i = 0; i < length; i++) {

      Value element = convertObjectToValue(Array.get(array, i));
      value.addElement(element);
    }

    return value;
  }



  private Value convertArrayArrayToValue (Object array, String elementType)
   throws Exception {

    int length = Array.getLength(array);
    ArrayValue value = new ArrayValue(elementType, length);

    for (int i = 0; i < length; i++) {

      Object object = Array.get(array, i);
      value.addElement(convertObjectToValue(object));
    }

    return value;
  }



  private Value convertReferenceArrayToValue (Object array, String elementType)
   throws Exception {

    int length = Array.getLength(array);
    ArrayValue value = new ArrayValue(elementType, length);

    for (int i = 0; i < length; i++) {

      Object object = Array.get(array, i);
      value.addElement(convertObjectToValue(object));
    }

    return value;
  }



  private Value convertReferenceToValue (Object object) throws Exception {
    
    Class c = object.getClass();
    ReferenceValue value = new ReferenceValue(c.getName());
    Enumeration e = getFields(c);
      
    while (e.hasMoreElements()) {

      Field field = (Field) e.nextElement();
      field.setAccessible(true);
      
      value.addValue(convertFieldToValue(field, object));
    }

    return value;
  }



  private Enumeration getFields (Class c) {

    Vector v = new Vector();

    while (c != null) {
      Field[] fields = c.getDeclaredFields();
      c = c.getSuperclass();

      for (int i = 0; i < fields.length; i++)
        v.addElement(fields[i]);
    }

    return v.elements();
  }



  private Value convertFieldToValue (Field field, Object object)
   throws Exception {

    Class type = field.getType();

    if (type.isPrimitive())
      return convertPrimitiveToValue(field, object);

    if (type == java.lang.String.class)
      return convertObjectToValue(field.get(object));

    if (type.isArray())
      return convertObjectToValue(field.get(object));

    return convertObjectToValue(field.get(object));
  }



  private Value convertPrimitiveToValue (Field field, Object object)
   throws Exception {

    Class type = field.getType();

    if (type == Boolean.TYPE)
      return new BooleanValue(field.getBoolean(object));

    if (type == Byte.TYPE)
      return new ByteValue(field.getByte(object));

    if (type == Character.TYPE)
      return new CharValue(field.getChar(object));

    if (type == Double.TYPE)
      return new DoubleValue(field.getDouble(object));

    if (type == Float.TYPE)
      return new FloatValue(field.getFloat(object));

    if (type == Integer.TYPE)
      return new IntValue(field.getInt(object));

    if (type == Long.TYPE)
      return new LongValue(field.getLong(object));

    if (type == Short.TYPE)
      return new ShortValue(field.getShort(object));

    throw new ValueConversionException("Unknown primitive type: " + type);
  }



  private void convertDummyRootValue (Value v, Value root) {

    if (v instanceof ReferenceValue)
      convertDummyRootValueInReference((ReferenceValue) v, root);
    else if (v instanceof ArrayValue)
      convertDummyRootValueInArray((ArrayValue) v, root);
  }



  private void convertDummyRootValueInReference (ReferenceValue rv,
   Value root) {

    int size = rv.getSize();

    for (int i = 0; i < size; i++) {

      Value tempValue = rv.getValueAt(i);
      
      if (tempValue == dummyRootValue)
        rv.setValueAt(i, root);
      else
        convertDummyRootValue(tempValue, root);
    }
  }



  private void convertDummyRootValueInArray (ArrayValue av, Value root) {

    int size = av.getSize();

    for (int i = 0; i < size; i++) {

      Value tempValue = av.getValueAt(i);
      
      if (tempValue == dummyRootValue)
        av.setValueAt(i, root);
      else
        convertDummyRootValue(tempValue, root);
    }
  }
}

/**
 * This class represents a generic reference. Reference objects are
 * the basic building blocks which form object graphs on the server
 * side from relational data. Reference values store a collection
 * of other values, which can be primitive, references or arrays.
 */

class ReferenceValue implements Value {

  private String className;
  private Vector values;

  public ReferenceValue () {

    className = null;
    values = new Vector();
  }

  public ReferenceValue (String className) {

    this.className = className;
    values = new Vector();
  }

  public String getClassName () {
    return className;
  }

  public String getSignature () {

    if (values == null)
      return "";

    StringBuffer buffer = new StringBuffer();
    Enumeration e = values.elements();

    while (e.hasMoreElements()) {

      Value v = (Value) e.nextElement();

      if (v instanceof ArrayValue)
        buffer.append(ARRAY);

      else if (v instanceof BooleanValue)
        buffer.append(BOOLEAN);

      else if (v instanceof CharValue)
        buffer.append(CHAR);

      else if (v instanceof DoubleValue)
        buffer.append(DOUBLE);

      else if (v instanceof FloatValue)
        buffer.append(FLOAT);

      else if (v instanceof IntValue)
        buffer.append(INT);

      else if (v instanceof LongValue)
        buffer.append(LONG);

      else if (v instanceof NullValue)
        buffer.append(NULL);

      else if (v instanceof ReferenceValue)
        buffer.append(REFERENCE);

      else if (v instanceof ShortValue)
        buffer.append(SHORT);

      else if (v instanceof StringValue)
        buffer.append(STRING);
    }

    return buffer.toString();
  }

  public void addValue (Value v) {

    if (values != null)
      values.addElement(v);
  }

  public int getSize () {

    if (values == null)
      return 0;
    else
      return values.size();
  }

  public Value getValueAt (int index) {

    if (values == null)
      return null;

    if (index >= 0 && index < values.size())
      return (Value) values.elementAt(index);
    else
      return null;
  }

  public void setValueAt (int index, Value newValue) {

    if (values == null || index < 0)
      return;

    values.setElementAt(newValue, index);
  }

  public Enumeration getValues () {

    if (values == null)
      return null;
    else
      return values.elements();
  }

  public String toString () {

    if (values == null) {
      return className + "{}";
    }

    StringBuffer buffer = new StringBuffer();

    buffer.append(className);
    buffer.append('{');

    Enumeration e = values.elements();

    if (e.hasMoreElements())
      buffer.append(e.nextElement());

    while (e.hasMoreElements()) {

      buffer.append(',');
      buffer.append(e.nextElement());
    }

    buffer.append('}');

    return buffer.toString();
  }
}

/**
 * This class represents a primitive short value.
 */

class ShortValue implements Value {

  private short value;

  public ShortValue () {
  }

  public ShortValue (short value) {
    this.value = value;
  }

  public short shortValue () {
    return value;
  }
  
  public String toString () {
    return String.valueOf(value);
  }
}

/**
 * This class represents a string value, although Java handles strings
 * as objects, databases handle strings as primitive values.
 */

class StringValue implements Value {

  private String value;

  public StringValue () {
    value = null;
  }

  public StringValue (String value) {
    this.value = value;
  }

  public String stringValue () {
    return value;
  }
  
  public String toString () {

    if (value == null)
      return "\"\"";
    else
      return  "\"" + value + "\"";
  }
}

/**
 */

interface Value extends ValueCodes, Serializable {
}


public class ValueConverter implements ValueCodes {



  private Hashtable refs;
  private Object object;
  private Value rootValue;
  private Hashtable rootObjectFields;
  private Vector rootObjectArrayElements;



  public ValueConverter (Value value) throws ValueConversionException {

    refs = new Hashtable();
    object = null;
    rootValue = value;
    rootObjectFields = new Hashtable();
    rootObjectArrayElements = new Vector();

    if (value == null)
      throw new ValueConversionException("Cannot convert null values");

    object = convertValueToObject(value);

    convertRootObjectFields(object);
  }



  public Object getObject () {
    return object;
  }



  private Object convertValueToObject (Value value)
   throws ValueConversionException {

    if (value instanceof NullValue)
      return null;

    if (value instanceof ArrayValue)
      return convertArrayToObject((ArrayValue) value);

    if (value instanceof ReferenceValue)
      return convertReferenceToObject((ReferenceValue) value);

    return convertPrimitiveToObject(value);
  }



  private Object convertArrayToObject (ArrayValue value)
   throws ValueConversionException {

    if (refs.containsKey(value))
      return refs.get(value);

    try {

      Class elementType = Class.forName(value.getElementType());
      Object obj = null;

      if (elementType.isPrimitive())
        obj = convertPrimitiveArrayToObject(value, elementType);

      else if (elementType.isArray())
        obj = convertArrayArrayToObject(value, elementType);

      else
        obj = convertReferenceArrayToObject(value, elementType);

      if (obj != null)
        refs.put(value, obj);

      return obj;

    } catch (Exception e) {

      throw new ValueConversionException();
    }
  }



  private Object convertPrimitiveArrayToObject (ArrayValue value,
   Class elementType) throws Exception {

    if (elementType == Boolean.TYPE)
      return convertBooleanArrayToObject(value);

    if (elementType == Byte.TYPE)
      return convertByteArrayToObject(value);

    if (elementType == Character.TYPE)
      return convertCharArrayToObject(value);

    if (elementType == Double.TYPE)
      return convertDoubleArrayToObject(value);

    if (elementType == Float.TYPE)
      return convertFloatArrayToObject(value);

    if (elementType == Integer.TYPE)
      return convertIntArrayToObject(value);

    if (elementType == Long.TYPE)
      return convertLongArrayToObject(value);

    if (elementType == Short.TYPE)
      return convertShortArrayToObject(value);

    throw new ValueConversionException(
     "Unknown array component type: " + elementType);
  }



  private Object convertBooleanArrayToObject (ArrayValue value)
   throws ValueConversionException {

    int length = value.getLength();
    Enumeration elements = value.getElements();
    boolean[] object = new boolean[length];

    for (int i = 0; i < length; i++) {

      Object next = elements.nextElement();

      if (next instanceof BooleanValue)
        object[i] = ((BooleanValue) next).booleanValue();
      else
        throw new ValueConversionException("Illegal type in boolean array");
    }

    return object;
  }



  private Object convertByteArrayToObject (ArrayValue value)
   throws ValueConversionException {

    int length = value.getLength();
    Enumeration elements = value.getElements();
    byte[] object = new byte[length];

    for (int i = 0; i < length; i++) {

      Object next = elements.nextElement();

      if (next instanceof ByteValue)
        object[i] = ((ByteValue) next).byteValue();
      else
        throw new ValueConversionException("Illegal type in byte array");
    }

    return object;
  }



  private Object convertCharArrayToObject (ArrayValue value)
   throws ValueConversionException {

    int length = value.getLength();
    Enumeration elements = value.getElements();
    char[] object = new char[length];

    for (int i = 0; i < length; i++) {

      Object next = elements.nextElement();

      if (next instanceof CharValue)
        object[i] = ((CharValue) next).charValue();
      else
        throw new ValueConversionException("Illegal type in char array");
    }

    return object;
  }



  private Object convertDoubleArrayToObject (ArrayValue value)
   throws ValueConversionException {

    int length = value.getLength();
    Enumeration elements = value.getElements();
    double[] object = new double[length];

    for (int i = 0; i < length; i++) {

      Object next = elements.nextElement();

      if (next instanceof DoubleValue)
        object[i] = ((DoubleValue) next).doubleValue();
      else
        throw new ValueConversionException("Illegal type in double array");
    }

    return object;
  }



  private Object convertFloatArrayToObject (ArrayValue value)
   throws ValueConversionException {

    int length = value.getLength();
    Enumeration elements = value.getElements();
    float[] object = new float[length];

    for (int i = 0; i < length; i++) {

      Object next = elements.nextElement();

      if (next instanceof FloatValue)
        object[i] = ((FloatValue) next).floatValue();
      else
        throw new ValueConversionException("Illegal type in float array");
    }

    return object;
  }



  private Object convertIntArrayToObject (ArrayValue value)
   throws ValueConversionException {

    int length = value.getLength();
    Enumeration elements = value.getElements();
    int[] object = new int[length];

    for (int i = 0; i < length; i++) {

      Object next = elements.nextElement();

      if (next instanceof IntValue)
        object[i] = ((IntValue) next).intValue();
      else
        throw new ValueConversionException("Illegal type in int array");
    }

    return object;
  }



  private Object convertLongArrayToObject (ArrayValue value)
   throws ValueConversionException {

    int length = value.getLength();
    Enumeration elements = value.getElements();
    long[] object = new long[length];

    for (int i = 0; i < length; i++) {

      Object next = elements.nextElement();

      if (next instanceof LongValue)
        object[i] = ((LongValue) next).longValue();
      else
        throw new ValueConversionException("Illegal type in long array");
    }

    return object;
  }



  private Object convertShortArrayToObject (ArrayValue value)
   throws ValueConversionException {

    int length = value.getLength();
    Enumeration elements = value.getElements();
    short[] object = new short[length];

    for (int i = 0; i < length; i++) {

      Object next = elements.nextElement();

      if (next instanceof ShortValue)
        object[i] = ((ShortValue) next).shortValue();
      else
        throw new ValueConversionException("Illegal type in short array");
    }

    return object;
  }



  private Object convertArrayArrayToObject (ArrayValue value,
   Class elementType) throws ValueConversionException {

    int length = value.getLength();
    Class type = getClass(value.getElementType());

    Object object = Array.newInstance(type, length);
    Enumeration elements = value.getElements();

    for (int i = 0; i < length; i++) {

      Object next = elements.nextElement();

      if (next instanceof NullValue) {

        Array.set(object, i, null);

      } else if (next instanceof ArrayValue) {

        Object ref = convertArrayToObject((ArrayValue) next);
        Array.set(object, i, ref);

      } else {

        throw new ValueConversionException("Illegal type in reference array");
      }
    }

    return object;
  }



  private Object convertReferenceArrayToObject (ArrayValue value,
   Class elementType) throws ValueConversionException {

    if (elementType == java.lang.String.class)
      return convertStringArrayToObject(value);

    int length = value.getLength();
    Class type = getClass(value.getElementType());

    Object object = Array.newInstance(type, length);
    Enumeration elements = value.getElements();

    for (int i = 0; i < length; i++) {

      Object next = elements.nextElement();

      if (next == rootValue) {

        rootObjectArrayElements.addElement(new ArrayElement(object, i));

      } else if (next instanceof NullValue) {

        Array.set(object, i, null);

      } else if (next instanceof ReferenceValue) {

        Object ref = convertReferenceToObject((ReferenceValue) next);
        Array.set(object, i, ref);

      } else {

        throw new ValueConversionException("Illegal type in reference array");
      }
    }

    return object;
  }



  private Object convertStringArrayToObject (ArrayValue value)
   throws ValueConversionException {

    int length = value.getLength();
    Enumeration elements = value.getElements();
    String[] object = new String[length];

    for (int i = 0; i < length; i++) {

      Object next = elements.nextElement();

      if (next instanceof StringValue)
        object[i] = ((StringValue) next).stringValue();
      else
        throw new ValueConversionException("Illegal type in string array");
    }

    return object;
  }



  private Class getClass (String type) throws ValueConversionException {

    try {

      return Class.forName(type);

    } catch (Exception e) {

      throw new ValueConversionException("Unknown type: " + type);
    }
  }



  private Object convertReferenceToObject (ReferenceValue refValue)
   throws ValueConversionException {

    if (refs.containsKey(refValue))
      return refs.get(refValue);

    Object object = null;

    try {

      Class type = getClass(refValue.getClassName());
      object = type.newInstance();

      Enumeration values = refValue.getValues();
      Enumeration fields = getFields(type);

      while (values.hasMoreElements() && fields.hasMoreElements()) {

        Value value = (Value) values.nextElement();
        Field field = (Field) fields.nextElement();
        Class fieldType = field.getType();

        field.setAccessible(true);

        if (value == rootValue) {

          rootObjectFields.put(object, field);
          continue;
        }

        if (value instanceof NullValue)
          field.set(object, null);

        else if (fieldType.isPrimitive())
          setFieldToPrimitiveValue(field, object, value);

        else if (fieldType.isArray())
          setFieldToArrayValue(field, object, value);

        else
          setFieldToReferenceValue(field, object, value);
      } 

    } catch (Exception e) {
    
      e.printStackTrace();
      throw new ValueConversionException();
    }

    refs.put(refValue, object);

    return object;
  }



  private void setFieldToPrimitiveValue (Field field, Object object,
   Value value) throws Exception {

    if (value instanceof BooleanValue) {

      BooleanValue bv = (BooleanValue) value;
      field.setBoolean(object, bv.booleanValue());

    } else if (value instanceof ByteValue) {

      ByteValue bv = (ByteValue) value;
      field.setByte(object, bv.byteValue());

    } else if (value instanceof CharValue) {

      CharValue cv = (CharValue) value;
      field.setChar(object, cv.charValue());

    } else if (value instanceof DoubleValue) {

      DoubleValue dv = (DoubleValue) value;
      field.setDouble(object, dv.doubleValue());

    } else if (value instanceof FloatValue) {

      FloatValue fv = (FloatValue) value;
      field.setFloat(object, fv.floatValue());

    } else if (value instanceof IntValue) {

      IntValue iv = (IntValue) value;
      field.setInt(object, iv.intValue());

    } else if (value instanceof LongValue) {

      LongValue lv = (LongValue) value;
      field.setLong(object, lv.longValue());

    } else if (value instanceof ShortValue) {

      ShortValue sv = (ShortValue) value;
      field.setShort(object, sv.shortValue());

    } else {

      throw new ValueConversionException("Unknown primitive value:" + value);
    }
  }



  private void setFieldToArrayValue (Field field, Object object,
   Value value) throws Exception {

    if (value instanceof ArrayValue) {

      ArrayValue arrayValue = (ArrayValue) value;
      field.set(object, convertArrayToObject(arrayValue));

    } else {

      throw new ValueConversionException("Value not an array: " + value);
    }
  }



  private void setFieldToReferenceValue (Field field, Object object,
   Value value) throws Exception {

    if (value instanceof StringValue) {

      StringValue strValue = (StringValue) value;
      field.set(object, strValue.stringValue());

    } else if (value instanceof ReferenceValue) {

      ReferenceValue refValue = (ReferenceValue) value;
      Object o = convertReferenceToObject(refValue);

      field.set(object, o);

    } else {

      throw new ValueConversionException("Value not a reference: " + value);
    }
  }



  private Enumeration getFields (Class c) {

    Vector v = new Vector();

    while (c != null) {

      Field[] fields = c.getDeclaredFields();
      c = c.getSuperclass();

      for (int i = 0; i < fields.length; i++)
        v.addElement(fields[i]);
    }

    return v.elements();
  }



  private Object convertPrimitiveToObject (Value value)
   throws ValueConversionException {

    if (value instanceof BooleanValue) {

      BooleanValue bv = (BooleanValue) value;
      return new Boolean(bv.booleanValue());

    } else if (value instanceof ByteValue) {

      ByteValue bv = (ByteValue) value;
      return new Byte(bv.byteValue());

    } else if (value instanceof CharValue) {

      CharValue cv = (CharValue) value;
      return new Character(cv.charValue());

    } else if (value instanceof DoubleValue) {

      DoubleValue dv = (DoubleValue) value;
      return new Double(dv.doubleValue());

    } else if (value instanceof FloatValue) {

      FloatValue fv = (FloatValue) value;
      return new Float(fv.floatValue());

    } else if (value instanceof IntValue) {

      IntValue iv = (IntValue) value;
      return new Integer(iv.intValue());

    } else if (value instanceof LongValue) {

      LongValue lv = (LongValue) value;
      return new Long(lv.longValue());

    } else if (value instanceof ShortValue) {

      ShortValue sv = (ShortValue) value;
      return new Short(sv.shortValue());

    } else if (value instanceof StringValue) {

      StringValue sv = (StringValue) value;
      return new String(sv.stringValue());

    }

    throw new ValueConversionException("Unknown primitive value: " + value);
  }



  private void convertRootObjectFields (Object root)
   throws ValueConversionException {

    try {

      Enumeration objects = rootObjectFields.keys();

      while (objects.hasMoreElements()) {

        Object o = objects.nextElement();
        Field f = (Field) rootObjectFields.get(o);

        f.set(o, root);
      }

      Enumeration elements = rootObjectArrayElements.elements();

      while (elements.hasMoreElements()) {

        ArrayElement ae = (ArrayElement) elements.nextElement();
        ae.setElement(root);
      }
    }

    catch (Exception e) {
    
      e.printStackTrace();
      throw new ValueConversionException(e.getMessage());
    }
  }
}

