 /**************************************************************************
/* This class is a simple JDBC test program.
/*
/* Copyright (c) 2001 by Bernhard Bablok (mail@bablokb.de)
/*
/* This program is free software; you can redistribute it and/or modify
/* it under the terms of the GNU General Public License as published
/* by  the Free Software Foundation; either version 2 of the License or
/* (at your option) any later version.
/*
/* This program is distributed in the hope that it will be useful, but
/* WITHOUT ANY WARRANTY; without even the implied warranty of
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/* GNU General Public License for more details.
/*
/* You should have received a copy of the GNU General Public License
/* along with this program; see the file COPYING.  If not, write to
/* the Free Software Foundation Inc., 59 Temple Place - Suite 330,
/* Boston, MA  02111-1307 USA
/**************************************************************************/
//package de.bablokb.jdbc;

import java.io.*;
import java.sql.*;

/**
   This class is a simple JDBC test program.

   @version $Revision: 1.15 $
   @author  $Author: Bablokb $
*/

public class DBTest {

  private static final String
    TABLE_CREATE  = "CREATE TABLE FileInfo (" + 
                    " filename VARCHAR(255) NOT NULL PRIMARY KEY," +
                    " size INT," +
                    " modtime TIMESTAMP," +
                    " username VARCHAR(8)," +
                    " groupname VARCHAR(8)," +
                    " links INT," +
                    " perms CHAR(4))",
    TABLE_DROP    = "DROP TABLE FileInfo",
    INSERT_HEADER = "INSERT INTO FileInfo" +
                    " (filename,size,modtime,username,groupname,links,perms) ",
    PREP_INSERT   = INSERT_HEADER + "VALUES(?,?,?,?,?,?,?)";

  private String iDriver, iUrl, iUser, iPassword;
  private Connection iConn;

  //////////////////////////////////////////////////////////////////////////////

  /**
     Constructor.
  */

  private DBTest() throws Exception {
    iDriver = System.getProperty("driver");
    if (iDriver == null)
      throw new Exception("Property 'driver' not specified!");
    iUrl = System.getProperty("url");
    if (iUrl == null)
      throw new Exception("Property 'url' not specified!");

    iUser = System.getProperty("user");
    iPassword = System.getProperty("password");

    Class.forName(iDriver);   // load database driver
    iConn = DriverManager.getConnection(iUrl,iUser,iPassword);
    if (iConn == null)
      throw new Exception("DriverManager.getConnection() returned null!");
    iConn.setAutoCommit(false);
  }

  //////////////////////////////////////////////////////////////////////////////

  public static void main(String[] args) {
    if (args.length == 0)
      usage();
    try {
      long start, stop, overhead=0;
      DBTest dbt = new DBTest();
      start = System.currentTimeMillis();
      if (args[0].equals("create"))
	dbt.createTable();
      else if (args[0].equals("import")) {
	if (args.length != 2)
	  usage();
	long ostart = System.currentTimeMillis();
	dbt.importOverhead(args[1]);
	overhead = System.currentTimeMillis() - ostart;
	dbt.importData(args[1]);
      } else if (args[0].equals("importp")) {
	if (args.length != 2)
	  usage();
	long ostart = System.currentTimeMillis();
	dbt.importOverhead(args[1]);
	overhead = System.currentTimeMillis() - ostart;
	dbt.importPrepared(args[1]);
      } else if (args[0].equals("search")) {
	if (args.length != 2)
	  usage();
	dbt.searchData(args[1]);
      } else if (args[0].equals("modify")) {
	if (args.length != 2)
	  usage();
	dbt.modifyData(args[1]);
      } else if (args[0].equals("drop"))
	dbt.dropTable();
      else
	usage();
      stop = System.currentTimeMillis();
      System.out.println("Total time: " + (stop-start-overhead) + " msec");
      dbt.close();
    } catch (Exception e) {
      e.printStackTrace();
      System.exit(3);
    }
  }

  //////////////////////////////////////////////////////////////////////////////

  /**
     Usage information
  */

  private static void usage() {
    System.out.println("usage: java " + DBTest.class.getName() + "options\n" +
       " Possible options:\n" +
       "   create\n" + 
       "   import <filename>\n" + 
       "   importp <filename>\n" + 
       "   search <clause>\n" + 
       "   modify <clause>\n" + 
       "   drop");
    System.exit(3);
  }

  //////////////////////////////////////////////////////////////////////////////

  /**
     Create table.
  */

  private void createTable() throws Exception {
    System.out.println("Creating table FileInfo in database " + iUrl);
    Statement s = iConn.createStatement();
    s.execute(TABLE_CREATE);    
    iConn.commit();
  }

  //////////////////////////////////////////////////////////////////////////////

  /**
     This method just does everything importData() does, except the database
     stuff. This method is used to calculate the overhead involved.
  */

  private void importOverhead(String source) throws Exception {
    System.out.println("Parsing import file for overhead calcuation...");
    BufferedReader stream = new BufferedReader(new FileReader(source));
    String line = stream.readLine();
    while (line != null) {
      FileInfo fi = new FileInfo(line);
      line = stream.readLine();
    }
    stream.close();
  }

  //////////////////////////////////////////////////////////////////////////////

  /**
     Import data into database.
  */

  private void importData(String source) throws Exception {
    System.out.println("Importing data to table FileInfo of database " + iUrl);

    Statement s = iConn.createStatement();
    BufferedReader stream = new BufferedReader(new FileReader(source));
    String line = stream.readLine();
    while (line != null) {
      FileInfo fi = new FileInfo(line);
      s.execute(INSERT_HEADER + fi.toInsertString());
      line = stream.readLine();
    }
    stream.close();
    iConn.commit();
  }

  //////////////////////////////////////////////////////////////////////////////

  /**
     Import data into database using PreparedStatement.
  */

  private void importPrepared(String source) throws Exception {
    System.out.println("Importing data to table FileInfo of database " + iUrl);

    PreparedStatement s = iConn.prepareStatement(PREP_INSERT);
    BufferedReader stream = new BufferedReader(new FileReader(source));
    String line = stream.readLine();
    while (line != null) {
      FileInfo fi = new FileInfo(line);
      fi.setParameter(s);
      s.execute();
      line = stream.readLine();
    }
    stream.close();
    iConn.commit();
  }

  //////////////////////////////////////////////////////////////////////////////

  /**
     Search (query) for specified data. Since we are only interested in
     timings, we don't analyse the ResultSet. This can be done in a 
     generic way using the ResultSetMetaData-object associated with the
     ResultSet.
  */

  private void searchData(String clause) throws Exception {
    System.out.println("Searching for data...");
    Statement s = iConn.createStatement();
    ResultSet rs = s.executeQuery(clause);
    rs.close();
  }

  //////////////////////////////////////////////////////////////////////////////

  /**
     Delete data from the database.
  */

  private void modifyData(String clause) throws Exception {
    System.out.println("Delete/Update of data from table FileInfo of  database "
		       + iUrl);
    Statement s = iConn.createStatement();
    int count = s.executeUpdate(clause);
    iConn.commit();
    System.out.println("Deleted/Updated " + count + " rows");
 }

  //////////////////////////////////////////////////////////////////////////////

  /**
     Drop table.
  */

  private void dropTable() throws Exception {
    System.out.println("Dropping table FileInfo in database " + iUrl);
    Statement s = iConn.createStatement();
    s.execute(TABLE_DROP);    
    iConn.commit();
  }

  //////////////////////////////////////////////////////////////////////////////

  /**
     Close connection.
  */

  private void close() throws SQLException {
    System.out.println("Closing connection to database " + iUrl);
    if (iConn != null && !iConn.isClosed())
      iConn.close();
  }
}
